#include "GLWindow.h"
#include <iostream>
#include "ngl/Camera.h"
#include "ngl/Transformation.h"
#include "ngl/TransformStack.h"
#include "ngl/NGLInit.h"
#include "ngl/ShaderManager.h"
#include "ngl/Util.h"
#include <QImage>
#include <ctime>

/// On the mac these are extentions, it's easier to let the
/// pre-processor deal with this than using GLEW / other code
/// as on Win and Linux it works as normal glGen

#ifdef DARWIN
  #define glGenVertexArrays glGenVertexArraysAPPLE
  #define glBindVertexArray glBindVertexArrayAPPLE
#endif


//----------------------------------------------------------------------------------------------------------------------
GLWindow::GLWindow(
                   QWidget *_parent
                  ) :
                    QGLWidget(_parent)
{

  // set this widget to have the initial keyboard focus
  setFocus();
  // re-size the widget to that of the parent (in this case the GLFrame passed in on construction)
  this->resize(_parent->size());
  // Now set the initial GLWindow attributes to default values
  // Roate is false
  m_rotate=false;
  // mouse rotation values set to 0
  m_spinXFace=0;
  m_spinYFace=0;
  m_origX=0;
  m_origY=0;
	m_fpsTimer =startTimer(0);
	m_fps=0;
	m_frames=0;
	m_timer.start();
	m_normalSize=0.1;
}

GLWindow::~GLWindow()
{
  ngl::NGLInit *Init = ngl::NGLInit::instance();
  std::cout<<"Shutting down NGL, removing VBO's and Shaders\n";
  Init->NGLQuit();
}




// This virtual function is called once before the first call to paintGL() or resizeGL(),
//and then once whenever the widget has been assigned a new QGLContext.
// This function should set up any required OpenGL context rendering flags, defining display lists, etc.

//----------------------------------------------------------------------------------------------------------------------
void GLWindow::initializeGL()
{

  glClearColor(0.4f, 0.4f, 0.4f, 1.0f);			   // Grey Background
  // enable depth testing for drawing
   glEnable(GL_DEPTH_TEST);
   // Now we will create a basic Camera from the graphics library
   // This is a static camera so it only needs to be set once
   // First create Values for the camera position
   ngl::Vector From(1,1,0.2);
   ngl::Vector To(0,0,0);
   ngl::Vector Up(0,1,0);
   ngl::NGLInit *Init = ngl::NGLInit::instance();
   #ifdef WIN32
   	glewInit();
   #endif
   Init->initGlew();
   m_cam= new ngl::Camera(From,To,Up,ngl::PERSPECTIVE);
   // set the shape using FOV 45 Aspect Ratio based on Width and Height
   // The final two are near and far clipping planes of 0.5 and 10
   m_cam->setShape(45,(float)720.0/576.0,0.5,150,ngl::PERSPECTIVE);
   // now to load the shader and set the values
   // grab an instance of shader manager
   ngl::ShaderManager *shader=ngl::ShaderManager::instance();
   // load a frag and vert shaders

	 shader->createShaderProgram("blinn");

	 shader->attachShader("blinnVertex",ngl::VERTEX);
	 shader->attachShader("blinnFragment",ngl::FRAGMENT);
	 shader->loadShaderSource("blinnVertex","shaders/Vertex.vs");
	 shader->loadShaderSource("blinnFragment","shaders/Fragment.fs");

	 shader->compileShader("blinnVertex");
	 shader->compileShader("blinnFragment");
	 shader->attachShaderToProgram("blinn","blinnVertex");
	 shader->attachShaderToProgram("blinn","blinnFragment");

	 shader->linkProgramObject("blinn");
	 shader->use("blinn");
	 ngl::NGLCheckGLError("in init",__LINE__);
   // now pass the modelView and projection values to the shader
   shader->setShaderParamFromMatrix("blinn","ViewMatrix",m_cam->getModelView());
   shader->setShaderParamFromMatrix("blinn","projectionMatrix",m_cam->getProjection());
	 shader->setShaderParam1i("blinn","Normalize",true);
	 shader->setShaderParam1i("blinn","numLightsEnabled",1);



	shader->createShaderProgram("normalShader");

	shader->attachShader("normalVertex",ngl::VERTEX);
	shader->attachShader("normalFragment",ngl::FRAGMENT);
	shader->loadShaderSource("normalVertex","shaders/normalVertex.vs");
	shader->loadShaderSource("normalFragment","shaders/normalFragment.fs");

	shader->compileShader("normalVertex");
	shader->compileShader("normalFragment");
	shader->attachShaderToProgram("normalShader","normalVertex");
	shader->attachShaderToProgram("normalShader","normalFragment");

	shader->attachShader("normalGeo",ngl::GEOMETRY);
	shader->loadShaderSource("normalGeo","shaders/normalGeo.gs");
	shader->compileShader("normalGeo");
	shader->attachShaderToProgram("normalShader","normalGeo");

	GLuint id=shader->getProgramID("normalShader");
	std::cout<<"shader id "<<id<<"\n";
	glProgramParameteriEXT(id,GL_GEOMETRY_VERTICES_OUT_EXT,8);
	glProgramParameteriEXT(id,GL_GEOMETRY_INPUT_TYPE_EXT,GL_TRIANGLES);
	glProgramParameteriEXT(id,GL_GEOMETRY_OUTPUT_TYPE_EXT,GL_LINE_STRIP);

	shader->linkProgramObject("normalShader");
	shader->use("normalShader");
	ngl::NGLCheckGLError("in init",__LINE__);
  // now pass the modelView and projection values to the shader
  shader->setShaderParamFromMatrix("normalShader","ViewMatrix",m_cam->getModelView());
  shader->setShaderParamFromMatrix("normalShader","projectionMatrix",m_cam->getProjection());
	shader->setShaderParam1f("normalShader","normalSize",0.1);
	shader->setShaderParam4f("normalShader","vertNormalColour",1,1,0,1);
	shader->setShaderParam4f("normalShader","faceNormalColour",1,0,0,1);

	shader->setShaderParam1i("normalShader","drawFaceNormals",true);
	shader->setShaderParam1i("normalShader","drawVertexNormals",true);

	glEnable(GL_DEPTH_TEST); // for removal of hidden surfaces
	ngl::VBOPrimitives *prim=ngl::VBOPrimitives::instance();
	prim->createVBOSphere("sphere",0.8,40);
}

//----------------------------------------------------------------------------------------------------------------------
//This virtual function is called whenever the widget has been resized.
// The new size is passed in width and height.
void GLWindow::resizeGL(
                        int _w,
                        int _h
                       )
{
  glViewport(0,0,_w,_h);
  m_cam->setShape(45,(float)_w/_h,0.5,150,ngl::PERSPECTIVE);
  ngl::ShaderManager *shader=ngl::ShaderManager::instance();
	shader->use("normalShader");
  shader->setShaderParamFromMatrix("normalShader","projectionMatrix",m_cam->getProjection());
	shader->use("blinn");
  shader->setShaderParamFromMatrix("blinn","projectionMatrix",m_cam->getProjection());

}

//----------------------------------------------------------------------------------------------------------------------
//This virtual function is called whenever the widget needs to be painted.
// this is our main drawing routine
void GLWindow::paintGL()
{
  // clear the screen and depth buffer
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
 // build our transform stack
  ngl::Transformation trans;
  // set the mouse rotation
  trans.setRotation(m_spinXFace,m_spinYFace,0);
  // set this in the TX stack
  m_transformStack.setGlobal(trans);

  ngl::ShaderManager *shader=ngl::ShaderManager::instance();
	ngl::VBOPrimitives *prim=ngl::VBOPrimitives::instance();

	(*shader)["normalShader"]->use();
  m_transformStack.loadGlobalAndCurrentMatrixToShader("normalShader","ModelMatrix");
	shader->setShaderParam1f("normalShader","normalSize",m_normalSize);

	prim->draw("teapot");


	(*shader)["blinn"]->use();

  m_transformStack.loadGlobalAndCurrentMatrixToShader("blinn","ModelMatrix");
	prim->draw("teapot");

}

//----------------------------------------------------------------------------------------------------------------------
void GLWindow::mouseMoveEvent (
                               QMouseEvent * _event
                              )
{
  // note the method buttons() is the button state when event was called
  // this is different from button() which is used to check which button was
  // pressed when the mousePress/Release event is generated
  if(m_rotate && _event->buttons() == Qt::LeftButton)
  {
    m_spinYFace = ( m_spinYFace + (_event->x() - m_origX) ) % 360 ;
    m_spinXFace = ( m_spinXFace + (_event->y() - m_origY) ) % 360 ;
    m_origX = _event->x();
    m_origY = _event->y();
  }
  // re-draw GL
  updateGL();
}


//----------------------------------------------------------------------------------------------------------------------
void GLWindow::mousePressEvent (
                                QMouseEvent * _event
                               )
{
  // this method is called when the mouse button is pressed in this case we
  // store the value where the maouse was clicked (x,y) and set the Rotate flag to true
  if(_event->button() == Qt::LeftButton)
  {
    m_origX = _event->x();
    m_origY = _event->y();
    m_rotate =true;
  }
}

//----------------------------------------------------------------------------------------------------------------------
void GLWindow::mouseReleaseEvent (
                                  QMouseEvent * _event
                                 )
{
  // this event is called when the mouse button is released
  // we then set Rotate to false
  if (_event->button() == Qt::LeftButton)
  {
    m_rotate=false;
  }
}

void GLWindow::timerEvent(
                          QTimerEvent *_event
                         )
{
	if(_event->timerId() == m_fpsTimer)
	  {
	    if( m_timer.elapsed() > 1000.0)
	    {
	      m_fps=m_frames;
	      m_frames=0;
	      m_timer.restart();
	    }
	   }
	    // re-draw GL
	updateGL();
}

void GLWindow::keyEvent(
												 QKeyEvent *_event
												)
{
	static bool fn=true;
	static bool vn=true;
	ngl::ShaderManager *shader=ngl::ShaderManager::instance();

  switch (_event->key())
  {
		case Qt::Key_Plus : m_normalSize+=0.01; break;
		case Qt::Key_Minus : m_normalSize-=0.01; break;
		case Qt::Key_1 :
				fn^=true;
				(*shader)["normalShader"]->use();

				shader->setShaderParam1i("normalShader","drawFaceNormals",fn);

		break;
		case Qt::Key_2 :
				vn^=true;
				(*shader)["normalShader"]->use();

				shader->setShaderParam1i("normalShader","drawVertexNormals",vn);

		break;
	}

}

